# coding: utf-8

"""
Wrapper for /bin/sign from obs-sign package
"""

from subprocess import Popen, PIPE, SubprocessError
import os
import time

from packaging import version

from copr_common.request import SafeRequest
from copr_backend.helpers import get_redis_logger
from .exceptions import CoprSignError, CoprSignNoKeyError, \
    CoprKeygenRequestError


SIGN_BINARY = "/bin/sign"

def create_gpg_email(username, projectname, domain):
    """
    Creates canonical name_email to identify gpg key
    """

    return "{}#{}@copr.{}".format(username, projectname, domain)


def call_sign_bin(cmd, log):
    """
    Call /bin/sign and return (rc, stdout, stderr).  Re-try the call
    automatically upon certain failures (if that makes sense).
    """
    cmd_pretty = ' '.join(cmd)
    for attempt in [1, 2, 3]:
        log.info("Calling '%s' (attempt #%s)", cmd_pretty, attempt)
        try:
            handle = Popen(cmd, stdout=PIPE, stderr=PIPE, encoding="utf-8")
            stdout, stderr = handle.communicate()
        except (SubprocessError, OSError) as err:
            new_err = CoprSignError("Failed to invoke '{}'".format(cmd_pretty))
            raise new_err from err

        if handle.returncode != 0:
            log.warning("Command '%s' failed with: %s",
                        cmd_pretty, stderr.rstrip())
            sleeptime = 20
            log.warning("Going to sleep %ss and re-try.", sleeptime)
            time.sleep(sleeptime)
            continue
        break
    return handle.returncode, stdout, stderr


def get_pubkey(username, projectname, log, sign_domain, outfile=None):
    """
    Retrieves public key for user/project from signer host.

    :param sign_domain: the domain name of the sign key
    :param outfile: [optional] file to write obtained key
    :return: public keys

    :raises CoprSignError: failed to retrieve key, see error message
    :raises CoprSignNoKeyError: if there are no such user in keyring
    """
    usermail = create_gpg_email(username, projectname, sign_domain)
    cmd = [SIGN_BINARY, "-u", usermail, "-p"]

    returncode, stdout, stderr = call_sign_bin(cmd, log)
    if returncode != 0:
        if "unknown key:" in stderr:
            raise CoprSignNoKeyError(
                "There are no gpg keys for user {} in keyring".format(username),
                return_code=returncode,
                cmd=cmd, stdout=stdout, stderr=stderr)
        raise CoprSignError(
            msg="Failed to get user pubkey\n"
                "sign stdout: {}\n sign stderr: {}\n".format(stdout, stderr),
            return_code=returncode,
            cmd=cmd, stdout=stdout, stderr=stderr)

    if outfile:
        with open(outfile, "w") as handle:
            handle.write(stdout)

    return stdout


def _sign_one(path, email, hashtype, log):
    cmd = [SIGN_BINARY, "-4", "-h", hashtype, "-u", email, "-r", path]
    returncode, stdout, stderr = call_sign_bin(cmd, log)
    if returncode != 0:
        raise CoprSignError(
            msg="Failed to sign {} by user {}".format(path, email),
            return_code=returncode,
            cmd=cmd, stdout=stdout, stderr=stderr)
    return stdout, stderr


def gpg_hashtype_for_chroot(chroot, opts):
    """
    Given the chroot name (in "mock format", like "fedora-rawhide-x86_64")
    return the expected GPG hash type.
    """
    # pylint: disable=too-many-return-statements

    el_chroots = ["rhel", "epel", "centos", "oraclelinux"]

    parts = chroot.split("-")

    version_part = parts[-2]

    if opts.gently_gpg_sha256:
        # For a few weeks we would use the sha256 hash type only for EL8+.
        # This is a safety belt, in case of any failure we'll just re-sign
        # epel-8+ and not _all_ the package data on backend.
        if parts[0] in el_chroots:
            el_version = version_part
            if el_version in ["rawhide"]:
                return "sha256"
            if version.parse(el_version) > version.parse("7"):
                return "sha256"
        return "sha1"

    if parts[0] in el_chroots:
        chroot_version = version_part
        if chroot_version in ["rawhide"]:
            return "sha256"
        if version.parse(chroot_version) <= version.parse("4"):
            return "sha1"
    if parts[0] == "fedora" and parts[1].isnumeric():
        # Fedora 27 moved to RPM v2.14 with the OpenSSL backend.
        if int(parts[1]) < 27:
            return "sha1"
    # fallback to sha256
    return "sha256"


def sign_rpms_in_dir(username, projectname, path, chroot, opts, log):
    """
    Signs rpms using obs-signd.

    If some some pkgs failed to sign, entire build marked as failed,
    but we continue to try sign other pkgs.

    :param username: copr username
    :param projectname: copr projectname
    :param path: directory with rpms to be signed
    :param chroot: chroot name where we sign packages, affects the hash type
    :param Munch opts: backend config

    :type log: logging.Logger

    :raises: :py:class:`backend.exceptions.CoprSignError` failed to sign at least one package
    """

    rpm_list = [
        os.path.join(path, filename.name)
        for filename in os.scandir(path)
        if filename.name.endswith(".rpm")
    ]

    if not rpm_list:
        return

    hashtype = gpg_hashtype_for_chroot(chroot, opts)

    try:
        get_pubkey(username, projectname, log, opts.sign_domain)
    except CoprSignNoKeyError:
        create_user_keys(username, projectname, opts, try_indefinitely=True)

    errors = []  # tuples (rpm_filepath, exception)
    for rpm in rpm_list:
        try:
            _sign_one(rpm, create_gpg_email(username, projectname, opts.sign_domain),
                      hashtype, log)
            log.info("signed rpm: %s", rpm)

        except CoprSignError as e:
            log.exception("failed to sign rpm: %s", rpm)
            errors.append((rpm, e))

    if errors:
        raise CoprSignError("Rpm sign failed, affected rpms: {}"
                            .format([err[0] for err in errors]))


def create_user_keys(username, projectname, opts, try_indefinitely=False):
    """
    Generate a new key-pair at sign host

    :param username:
    :param projectname:
    :param opts: backend config

    :return: None
    """
    data = {
        "name_real": "{}_{}".format(username, projectname),
        "name_email": create_gpg_email(username, projectname, opts.sign_domain)
    }

    log = get_redis_logger(opts, "sign", "actions")
    keygen_url = "http://{}/gen_key".format(opts.keygen_host)
    query = dict(url=keygen_url, data=data, method="post")
    try:
        request = SafeRequest(log=log, try_indefinitely=try_indefinitely)
        response = request.send(**query)
    except Exception as e:
        raise CoprKeygenRequestError(
            msg="Failed to create key-pair for user: {},"
                " project:{} with error: {}"
            .format(username, projectname, e), request=query)

    if response.status_code >= 400:
        raise CoprKeygenRequestError(
            msg="Failed to create key-pair for user: {}, project:{}, status_code: {}, response: {}"
            .format(username, projectname, response.status_code, response.text),
            request=query, response=response)


def _unsign_one(path):
    # Requires rpm-sign package
    cmd = ["/usr/bin/rpm", "--delsign", path]
    handle = Popen(cmd, stdout=PIPE, stderr=PIPE, encoding="utf-8")
    stdout, stderr = handle.communicate()

    if handle.returncode != 0:
        err = CoprSignError(
            msg="Failed to unsign {}".format(path),
            return_code=handle.returncode,
            cmd=cmd, stdout=stdout, stderr=stderr)

        raise err

    return stdout, stderr


def unsign_rpms_in_dir(path, opts, log):
    """
    :param path: directory with rpms to be signed
    :param Munch opts: backend config
    :type log: logging.Logger
    :raises: :py:class:`backend.exceptions.CoprSignError` failed to sign at least one package
    """
    rpm_list = [
        os.path.join(path, filename.name)
        for filename in os.scandir(path)
        if filename.name.endswith(".rpm")
        ]

    if not rpm_list:
        return

    errors = []  # tuples (rpm_filepath, exception)
    for rpm in rpm_list:
        try:
            _unsign_one(rpm)
            log.info("unsigned rpm: %s", rpm)

        except CoprSignError as e:
            log.exception("failed to unsign rpm: %s", rpm)
            errors.append((rpm, e))

    if errors:
        raise CoprSignError("Rpm unsign failed, affected rpms: {}"
                            .format([err[0] for err in errors]))


def resign_rpms_in_dir(username, projectname, path, chroot, opts, log):
    """
    Drop old signatures coming from original repo and re-sign.
    """
    # pylint: disable=too-many-positional-arguments
    unsign_rpms_in_dir(path, opts=opts, log=log)
    if opts.do_sign:
        sign_rpms_in_dir(username, projectname, path, chroot, opts=opts, log=log)



"""
Shared logic for hitcounter scripts
"""

import os
import re
from datetime import datetime
from requests.utils import unquote
from copr_common.request import SafeRequest
from copr_backend.helpers import BackendConfigReader

base_regex = "/results/(?P<owner>[^/]*)/(?P<project>[^/]*)/(?P<chroot>[^/]*)/"
repomd_url_regex = re.compile(base_regex + "repodata/repomd.xml", re.IGNORECASE)
rpm_url_regex = re.compile(
    base_regex + r"(?P<build_dir>[^/]*|Packages/[a-z])/(?P<rpm>[^/]*\.rpm)",
    re.IGNORECASE,
)

spider_regex = re.compile(
    '.*(ahrefs|bot/[0-9]|bingbot|borg|google|googlebot|yahoo|slurp|msnbot'
    '|openbot|archiver|netresearch|lycos|scooter|altavista|teoma|gigabot'
    '|blitzbot|oegp|charlotte|furlbot|http://client|polybot|htdig|ichiro'
    '|larbin|pompos|scrubby|searchsight|seekbot|semanticdiscovery|silk|snappy'
    '|spider|voila|vortex|voyager|zao|zeal|fast-webcrawler|converacrawler'
    '|msrbot|baiduspider|mogimogi|speedy|dataparksearch'
    '|findlinks|crawler|yandex|blexbot|semrushbot).*',
    re.IGNORECASE)


def url_to_key_strings(url):
    """
    Take a full URL and return a list of unique strings representing it,
    that copr-frontend will understand.
    """
    url_match = repomd_url_regex.match(url)
    if url_match:
        chroot_key = (
            'chroot_repo_metadata_dl_stat',
            url_match.group('owner'),
            url_match.group('project'),
            url_match.group('chroot')
        )
        chroot_key_str = '|'.join(chroot_key)
        return [chroot_key_str]

    url_match = rpm_url_regex.match(url)
    if url_match:
        chroot_key = (
            'chroot_rpms_dl_stat',
            url_match.group('owner'),
            url_match.group('project'),
            url_match.group('chroot')
        )
        chroot_key_str = '|'.join(chroot_key)
        project_key = (
            'project_rpms_dl_stat',
            url_match.group('owner'),
            url_match.group('project')
        )
        project_key_str = '|'.join(project_key)
        return [chroot_key_str, project_key_str]
    return []


def update_frontend(accesses, log, dry_run=False, try_indefinitely=False):
    """
    Increment frontend statistics based on these `accesses`
    """
    result = get_hit_data(accesses, log)
    if not result:
        log.debug("No recognizable hits among these accesses, skipping.")
        return

    log.debug(
        "Sending: %i results from %i to %i",
        len(result["hits"]),
        result["ts_from"],
        result["ts_to"]
    )
    if len(result["hits"]) < 100:
        log.debug("Hits: %s", result["hits"])
    else:
        log.debug("Not logging the whole dict: %s hits", len(result["hits"]))

    opts = BackendConfigReader().read()
    url = os.path.join(
        opts.frontend_base_url,
        "stats_rcv",
        "from_backend",
    )
    if not dry_run:
        request = SafeRequest(
            auth=("user", opts.frontend_auth),
            log=log,
            try_indefinitely=try_indefinitely,
        )
        request.post(url, result)


def get_hit_data(accesses, log):
    """
    Prepare body for the frontend request in the same format that
    copr_log_hitcounter.py does.
    """
    hits = {}
    timestamps = []
    for access in accesses:
        url = access["cs-uri-stem"]

        if access["sc-status"] == "404":
            log.debug("Skipping: %s (404 Not Found)", url)
            continue

        if access["cs(User-Agent)"].startswith("Mock"):
            log.debug("Skipping: %s (user-agent: Mock)", url)
            continue

        bot = spider_regex.match(access["cs(User-Agent)"])
        if bot:
            log.debug("Skipping: %s (user-agent '%s' is a known bot)",
                      url, bot.group(1))
            continue

        # Convert encoded characters from their %40 values back to @.
        url = unquote(url)

        # I don't know how or why but occasionally there is an URL that is
        # encoded twice (%2540oamg -> %40oamg - > @oamg), and yet its status
        # code is 200. AFAIK these appear only for EPEL-7 chroots and their
        # User-Agent is something like urlgrabber/3.10%20yum/3.4.3
        # I wasn't able to reproduce such accesses, and we decided to not count
        # them
        if url != unquote(url):
            log.warning("Skipping: %s (double encoded URL, user-agent: '%s', "
                        "status: %s)", access["cs-uri-stem"],
                        access["cs(User-Agent)"], access["sc-status"])
            continue

        # We don't want to count every accessed URL, only those pointing to
        # RPM files and repo file
        key_strings = url_to_key_strings(url)
        if not key_strings:
            log.debug("Skipping: %s", url)
            continue

        if any(x for x in key_strings
               if x.startswith("chroot_rpms_dl_stat|")
               and x.endswith("|srpm-builds")):
            log.debug("Skipping %s (SRPM build)", url)
            continue

        log.debug("Processing: %s", url)

        # When counting RPM access, we want to iterate both project hits and
        # chroot hits. That way we can get multiple `key_strings` for one URL
        for key_str in key_strings:
            hits.setdefault(key_str, 0)
            hits[key_str] += 1

        # Remember this access timestamp
        datetime_format = "%Y-%m-%d %H:%M:%S"
        datetime_string = "{0} {1}".format(access["date"], access["time"])
        datetime_object = datetime.strptime(datetime_string, datetime_format)
        timestamps.append(int(datetime_object.timestamp()))

    return {
        "ts_from": min(timestamps),
        "ts_to": max(timestamps),
        "hits": hits,
    } if hits else {}



"""
ActionDispatcher related classes.
"""

from copr_backend.exceptions import FrontendClientException
from copr_backend.dispatcher import BackendDispatcher
from copr_backend.rpm_builds import BlockedOwnersLimit

from ..actions import ActionWorkerManager, ActionQueueTask, Action

class ActionDispatcher(BackendDispatcher):
    """
    Kick-off action dispatcher daemon.
    """
    task_type = 'action'
    worker_manager_class = ActionWorkerManager

    def __init__(self, backend_opts):
        super().__init__(backend_opts)
        self.max_workers = backend_opts.actions_max_workers

        blocked_owners = backend_opts.builds_limits["blocked_owners"]
        self.limits.append(BlockedOwnersLimit(blocked_owners))
        self.log.info("setting limit for blocked owners: %s", blocked_owners)

    def get_frontend_tasks(self):
        try:
            raw_actions = self.frontend_client.get('pending-actions').json()
        except (FrontendClientException, ValueError) as error:
            self.log.exception(
                "Retrieving an action tasks failed with error: %s",
                error)
            return []

        return [ActionQueueTask(Action(self.opts, action, log=self.log))
                for action in raw_actions]



"""
Test Pulp client
"""

# pylint: disable=attribute-defined-outside-init

from unittest.mock import Mock
from copr_backend.pulp import PulpClient


class TestPulp:

    def setup_method(self, _method):
        self.config = {
            "api_root": "/pulp/",
            "base_url": "http://pulp.fpo:24817",
            "cert": "",
            "domain": "default",
            "dry_run": False,
            "format": "json",
            "key": "",
            "password": "1234",
            "timeout": 0,
            "username": "admin",
            "verbose": 0,
            "verify_ssl": True,
        }

    def test_url(self):
        client = PulpClient(self.config)
        assert self.config["domain"] == "default"
        assert client.url("api/v3/artifacts/")\
            == "http://pulp.fpo:24817/pulp/api/v3/artifacts/"

        assert client.url("api/v3/repositories/rpm/rpm/?")\
            == "http://pulp.fpo:24817/pulp/api/v3/repositories/rpm/rpm/?"

        self.config["domain"] = "copr"
        assert client.url("api/v3/artifacts/")\
            == "http://pulp.fpo:24817/pulp/copr/api/v3/artifacts/"

    def create_mock_response(self, results, count, next_url=None, ok=True):
        mock_response = Mock()
        mock_response.ok = ok
        mock_response.status_code = 200 if ok else 400
        mock_response.text = "OK" if ok else "Error"
        mock_response.json.return_value = {
            "count": count,
            "next": next_url,
            "previous": None,
            "results": results
        }
        return mock_response

    def test_get_content_pagination_single_page(self):
        client = PulpClient(self.config)

        results = [{"prn": f"rpm-{i}"} for i in range(50)]
        mock_response = self.create_mock_response(results, 50, next_url=None)
        client.send = Mock(return_value=mock_response)

        response = client.get_content([1234], fields=["prn"])

        assert response.ok
        data = response.json()
        assert data["count"] == 50
        assert len(data["results"]) == 50
        assert data["next"] is None
        assert all(item["prn"] == f"rpm-{i}" for i, item in enumerate(data["results"]))

        assert client.send.call_count == 1

    def test_get_content_pagination_multiple_pages(self):
        client = PulpClient(self.config)

        def mock_send(_, uri):
            if "offset=0" in uri:
                # First page
                results = [{"prn": f"rpm-{i}"} for i in range(1000)]
                return self.create_mock_response(
                    results, 2500, next_url="http://test/api/v3/content/rpm/packages/?offset=1000"
                )
            if "offset=1000" in uri:
                # Second page
                results = [{"prn": f"rpm-{i}"} for i in range(1000, 2000)]
                return self.create_mock_response(
                    results, 2500, next_url="http://test/api/v3/content/rpm/packages/?offset=2000"
                )
            if "offset=2000" in uri:
                # Third page (partial)
                results = [{"prn": f"rpm-{i}"} for i in range(2000, 2500)]
                return self.create_mock_response(results, 2500, next_url=None)
            return self.create_mock_response([], 2500, next_url=None)

        client.send = Mock(side_effect=mock_send)

        response = client.get_content([1234, 5678], fields=["prn"])

        assert response.ok
        data = response.json()
        assert data["count"] == 2500
        assert len(data["results"]) == 2500
        assert data["next"] is None
        assert all(item["prn"] == f"rpm-{i}" for i, item in enumerate(data["results"]))

        assert client.send.call_count == 3

    def test_get_content_pagination_error_handling(self):
        client = PulpClient(self.config)
        error_response = self.create_mock_response([], 0, ok=False)
        client.send = Mock(return_value=error_response)
        response = client.get_content([1234])

        assert not response.ok
        assert client.send.call_count == 1



"""
BuildDispatcher related classes.
"""

from copr_common.worker_manager import HashWorkerLimit
from copr_backend.dispatcher import BackendDispatcher
from copr_backend.rpm_builds import (
    ArchitectureWorkerLimit,
    ArchitectureUserWorkerLimit,
    BuildTagLimit,
    UserSSHLimit,
    BlockedOwnersLimit,
    RPMBuildWorkerManager,
    BuildQueueTask,
)
from ..exceptions import FrontendClientException


class _PriorityCounter:
    def __init__(self):
        self._counter = {}

    def get_priority(self, task):
        """
        Calculate the "dynamic" task priority, based on the actual queue size
        and queue structure.  We have this matrix of counters:

            _counter["background"]["user"]["arch"]["sandbox"] = value

        Each dimension "re-sets" the priority and starts counting from zero.
        Simply, the later the task is submitted (higher build ID) the less
        priority it gets (== gets higher priority number).

        We want to have separate counters for:

        - normal and background builds (background jobs are always penalized by
          the BuildQueueTask.frontend_priority), those are two separate but
          equivalent queues here

        - for each architecture (includes the source builds), that's because
          each architecture needs to have it's own "queue"/pool of builders
          (doesn't make sense to penalize e.g. ppc64le tasks and delay them
          because of a heavy x86_64 queue)

        - For each sandbox.  Each sandbox is equivalent to each other.  This
          means that builds submitted to 'jdoe/foo--user1' should be taken with
          the same priority as 'jdoe/baz-user1' (note the same owner pattern!)
          no matter *when* they are submitted.  This is useful when 'jdoe/foo'
          get's huge batch of builds, and _then_ after some time some _other_
          builds are submitted into the 'jdoe/baz'; those 'jdoe/baz builds could
          be blocked for a very long time.  See below.

        Note that for large queues, build_dispatcher isn't sometimes able to
        handle all the tasks in the priority queue in one cycle (we are able to
        process several hundreds, or at most thousands of tasks — depending on
        the 'sleeptime' value).  Therefore, the priority really matters here, as
        we want to avoid all kinds of the weird builder/queue starving
        situations.  In other words — if task isn't processed in one
        WorkerManager.run(timeout=???) cycle — it may stay "pending" for many
        other cycles too (depending on how quickly the overall queue get's
        processed).
        """

        def _get_subdict(source, arg):
            source.setdefault(arg, {})
            return source[arg]

        background = _get_subdict(self._counter, task.background)
        owner = _get_subdict(background, task.owner)
        arch = _get_subdict(owner, task.requested_arch or "srpm")

        # calculate from zero
        arch.setdefault(task.sandbox, 0)
        arch[task.sandbox] += 1
        return arch[task.sandbox]


class BuildDispatcher(BackendDispatcher):
    """
    Kick-off build dispatcher daemon.
    """
    task_type = 'build'
    worker_manager_class = RPMBuildWorkerManager

    def __init__(self, backend_opts):
        super().__init__(backend_opts)
        self.max_workers = backend_opts.builds_max_workers

        for tag_type in ["arch", "tag", "arch_per_owner"]:
            match tag_type:
                case "arch":
                    lclass = ArchitectureWorkerLimit
                case "tag":
                    lclass = BuildTagLimit
                case "arch_per_owner":
                    lclass = ArchitectureUserWorkerLimit
            for tag, limit in backend_opts.builds_limits[tag_type].items():
                self.log.info("setting %s(%s) limit to %s", tag_type, tag, limit)
                self.limits.append(lclass(tag, limit))

        for limit_type in ['sandbox', 'owner']:
            max_builders = backend_opts.builds_limits[limit_type]
            self.log.info("setting %s limit to %s", limit_type, max_builders)
            self.limits.append(HashWorkerLimit(
                lambda x, limit=limit_type: getattr(x, limit),
                max_builders,
                name=limit_type,
            ))

        blocked_owners = backend_opts.builds_limits["blocked_owners"]
        self.limits.append(BlockedOwnersLimit(blocked_owners))
        self.log.info("setting limit for blocked owners: %s", blocked_owners)

        limit = backend_opts.builds_limits["userssh"]
        userssh = UserSSHLimit(limit)
        self.log.info("setting %s limit to %s", "userssh", limit)
        self.limits.append(userssh)

    def get_frontend_tasks(self):
        """
        Retrieve a list of build jobs to be done.
        """
        try:
            raw_tasks = self.frontend_client.get('pending-jobs').json()
        except (FrontendClientException, ValueError) as error:
            self.log.exception("Retrieving build jobs from %s failed with error: %s",
                               self.opts.frontend_base_url, error)
            return []

        tasks = []
        priority = _PriorityCounter()
        for raw in raw_tasks:
            task = BuildQueueTask(raw)
            task.backend_priority = priority.get_priority(task)
            tasks.append(task)
        return tasks

    def get_cancel_requests_ids(self):
        try:
            return self.frontend_client.get('build-tasks/cancel-requests').json()
        except (FrontendClientException, ValueError) as error:
            self.log.exception("Retrieving build jobs from %s failed with error: %s",
                               self.opts.frontend_base_url, error)
            return []

    def report_canceled_task_id(self, task_id, was_running):
        self.frontend_client.post('build-tasks/canceled/{}'.format(task_id),
                                  data=None if not was_running else was_running)



"""
The Copr project globals to simplify access to config, logging, etc.
"""

import logging
import os
from copr_backend.helpers import (
    BackendConfigReader,
    get_redis_log_handler,
    RedisPublishHandler,
)


class App:
    """
    Shortcut.
    """
    log = None

    def __init__(self):
        self._setup_logging()
        self.config_file = os.environ.get("BACKEND_CONFIG",
                                          "/etc/copr/copr-be.conf")
        self.opts = BackendConfigReader(self.config_file).read()

    def _setup_logging(self):
        logging.basicConfig(level=logging.DEBUG)
        self.log = logging.getLogger()
        self.log.setLevel(logging.DEBUG)

    def setup_redis_logging(self, filename):
        """
        Setup a multiprocessing logger to a file inside /var/log/copr-backend,
        using the Redis handler.
        """
        for handler in self.log.handlers:
            if isinstance(handler, RedisPublishHandler):
                return
        self.log.addHandler(get_redis_log_handler(self.opts, filename))

    def redirect_to_redis_log(self, filename):
        """
        Drop all handlers from self.log, and add one going through Redis
        """
        self.log.handlers = []
        self.setup_redis_logging(filename)



# coding: utf-8

from munch import Munch
from requests import Response

from copr_common.request import RequestRetryError
from copr_backend.frontend import FrontendClient
from copr_backend.exceptions import FrontendClientException

from unittest import mock
from unittest.mock import MagicMock
import pytest

@pytest.fixture
def post_req():
    with mock.patch("copr_common.request.requests.request") as obj:
        yield obj

@pytest.fixture(scope='function', params=['get', 'post', 'put'])
def f_request_method(request):
    'mock the requests.{get,post,put} method'
    with mock.patch("copr_common.request.requests.request") as ctx:
        ctx.return_value.headers = {
            "Copr-FE-BE-API-Version": "666",
        }
        yield (request.param, ctx)


@pytest.fixture
def mc_time():
    with mock.patch("copr_common.request.time") as obj:
        yield obj


class TestFrontendClient(object):

    @staticmethod
    def _get_fake_response():
        resp = Munch()
        resp.headers = {
            "Copr-FE-BE-API-Version": "666",
        }
        resp.status_code = 200
        resp.data = "ok\n"
        return resp

    def setup_method(self, method):
        self.opts = Munch(
            frontend_base_url="http://example.com/",
            frontend_auth="12345678",
        )
        self.fc = FrontendClient(self.opts)

        self.data = {
            "foo": "bar",
            "bar": [1, 3, 5],
        }
        self.url_path = "sub_path"

        self.build_id = 12345
        self.task_id = "12345-fedora-20-x86_64"
        self.chroot_name = "fedora-20-x86_64"

    @pytest.fixture
    def mask_frontend_request(self):
        with mock.patch("copr_common.request.SafeRequest._send_request") as obj:
            yield obj

    def test_post_to_frontend(self, f_request_method):
        name, method = f_request_method
        method.return_value.status_code = 200
        self.fc.send(self.url_path, method=name, data=self.data)
        assert method.called

    def test_post_to_frontend_wrappers(self, f_request_method):
        name, method = f_request_method
        method.return_value.status_code = 200

        call = getattr(self.fc, name)
        if name == 'get':
            call(self.url_path)
        else:
            call(self.url_path, self.data)

        assert method.called

    def test_post_to_frontend_repeated_first_try_ok(self, mask_frontend_request, mc_time):
        mc_time.time.return_value = 0
        response = self._get_fake_response()
        mask_frontend_request.return_value = response
        assert self.fc.post(self.data, self.url_path) == response
        assert not mc_time.sleep.called

    def test_post_to_frontend_repeated_second_try_ok(self, f_request_method,
            mask_frontend_request, mc_time):
        method_name, method = f_request_method

        response = self._get_fake_response()
        mask_frontend_request.side_effect = [
            RequestRetryError(),
            response,
        ]
        mc_time.time.return_value = 0
        assert self.fc.send(
            self.url_path,
            data=self.data,
            method=method_name
        ) == response
        assert mc_time.sleep.called

    def test_post_to_frontend_err_400(self, post_req, mc_time):
        response = Response()
        response.status_code = 404
        response.reason = 'NOT FOUND'

        post_req.side_effect = [
            RequestRetryError(),
            response,
        ]

        mc_time.time.return_value = 0
        with pytest.raises(FrontendClientException):
            assert self.fc.post(self.data, self.url_path) == response
        assert mc_time.sleep.called

    def test_post_to_frontend_repeated_all_attempts_failed(self,
            mask_frontend_request, caplog, mc_time):
        mc_time.time.side_effect = [0, 0, 5, 5+10, 5+10+15, 5+10+15+20, 1000]
        mask_frontend_request.side_effect = RequestRetryError()
        with pytest.raises(FrontendClientException):
            self.fc.post(self.data, self.url_path)
        assert mc_time.sleep.call_args_list == [mock.call(x) for x in [5, 10, 15, 20, 25]]
        records = [x for x in caplog.records if "Retry request" in x.msg]
        assert len(records) == 5

    def test_post_to_frontend_repeated_indefinitely(self,
            mask_frontend_request, caplog, mc_time):
        mc_time.time.return_value = 1
        self.fc.try_indefinitely = True
        mask_frontend_request.side_effect = [RequestRetryError() for _ in range(100)] \
                             + [FrontendClientException()] # e.g. 501 eventually
        with pytest.raises(FrontendClientException):
            self.fc.post(self.data, self.url_path)
        assert mc_time.sleep.called
        records = [x for x in caplog.records if "Retry request" in x.msg]
        assert len(records) == 100

    def test_retries_on_outdated_frontend(self, mask_frontend_request, caplog):
        response = self._get_fake_response()
        response.headers["Copr-FE-BE-API-Version"] = "0"
        mask_frontend_request.side_effect = [
            response for _ in range(100)] + [Exception("sorry")]
        with pytest.raises(Exception):
            self.fc.try_indefinitely = True
            self.fc.post(self.url_path, self.data)
        assert len(mask_frontend_request.call_args_list) == 101
        assert "Sending POST request to frontend" in caplog.records[0].getMessage()
        assert "Copr FE/BE API is too old on Frontend" in caplog.records[1].msg

    def test_update(self):
        ptfr = MagicMock()
        self.fc.post = ptfr
        self.fc.update(self.data)
        assert ptfr.call_args == mock.call("update", self.data)

    def test_starting_build(self):
        ptfr = MagicMock()
        self.fc.post = ptfr
        for val in [True, False]:
            ptfr.return_value.json.return_value = {"can_start": val}

            assert self.fc.starting_build(self.data) == val

    def test_starting_build_err(self):
        ptfr = MagicMock()
        self.fc.post = ptfr

        with pytest.raises(FrontendClientException):
            self.fc.starting_build(self.data)

    def test_starting_build_err_2(self):
        ptfr = MagicMock()
        self.fc.post = ptfr
        ptfr.return_value.json.return_value = {}

        with pytest.raises(FrontendClientException):
            self.fc.starting_build(self.data)

    def test_reschedule_build(self):
        ptfr = MagicMock()
        self.fc.post = ptfr
        self.fc.reschedule_build(self.build_id, self.task_id, self.chroot_name)
        expected = mock.call('reschedule_build_chroot', {
            'build_id': self.build_id,
            'task_id': self.task_id,
            'chroot': self.chroot_name,
        })
        assert ptfr.call_args == expected



"""
Pulp doesn't provide an API client, we are implementing it for ourselves
"""

import io
import json
import logging
import os
import time
import tomllib
import hashlib
from urllib.parse import urlencode

from copr_common.request import SafeRequest
from copr_common.lock import lock, LockTimeout
from copr_common.redis_helpers import get_redis_connection


class PaginatedResponse:
    """
    A response-like object that maintains backward compatibility with PULP API responses
    while containing all paginated results.
    """
    def __init__(self, all_results, original_response):
        self._all_results = all_results
        self._original_response = original_response

        self.status_code = original_response.status_code
        self.ok = original_response.ok
        self.text = original_response.text

    def json(self):
        """
        Mimic Pulp paginated response, but with all results included
        """
        return {
            "count": len(self._all_results),
            "next": None,  # All results are included
            "previous": None,
            "results": self._all_results
        }

    def raise_for_status(self):
        """
        Mimic requests.Response.raise_for_status()
        """
        self._original_response.raise_for_status()


class BatchedAddRemoveContent:
    """
    Group a set of `add_and_remove` tasks for a single Pulp repository.
    """
    def __init__(self, repo_href, rpms_to_add, rpms_to_remove, backend_opts=None, log=None, noop=False):
        self.repo_href = repo_href
        self.rpms_to_add = rpms_to_add or []
        self.rpms_to_remove = rpms_to_remove or []
        self.log = log
        self.noop = noop

        if not backend_opts:
            self.log.error("can't get access to redis, batch disabled")
            self.noop = True
            return

        self._pid = os.getpid()
        self._json_redis_task = json.dumps({
            "add_content_units": self.rpms_to_add,
            "remove_content_units": self.rpms_to_remove,
        })

        self.notify_keys = []
        self.redis = get_redis_connection(backend_opts)

    @property
    def key(self):
        """ Our instance ID (key in Redis DB) """
        return "add_remove_batched::{}::{}".format(
            self.repo_href, self._pid)

    @property
    def key_pattern(self):
        """ Redis key pattern for potential tasks we can batch-process """
        return "add_remove_batched::{}::*".format(self.repo_href)

    def make_request(self):
        """ Request the task into Redis DB.  Run _before_ lock! """
        if self.noop:
            return None
        self.redis.hset(self.key, "task", self._json_redis_task)
        return self.key

    def check_processed(self, delete_if_not=True):
        """
        Drop our entry from Redis DB (if any), and return True if the task is
        already processed.  When 'delete_if_not=True, we delete the self.key
        from Redis even if the task is not yet processed (meaning that caller
        plans to finish the task right away).
        """
        if self.noop:
            return False

        status = self.redis.hget(self.key, "status") == "success"
        self.log.debug("Has already a status? %s", status)

        try:
            if not status:
                # not yet processed
                return False
        finally:
            # This is atomic operation, other processes may not re-start doing this
            # task again.  https://github.com/redis/redis/issues/9531
            #
            # The 'delete_if_not = True' is not a TOCTOU case.  Our caller holds
            # the lock, hence no other machine may take our task (would require
            # lock) and re-process it.
            if status or delete_if_not:
                self.redis.delete(self.key)

        return status

    def options(self):
        """
        Get the options from other _compatible_ (see below) Redis tasks, and
        plan the list of tasks in self.notify_keys[] that we will notify in
        commit().
        """
        rpms_to_add = set(self.rpms_to_add)
        rpms_to_remove = set(self.rpms_to_remove)

        if self.noop:
            return {"add_content_units": rpms_to_add, "remove_content_units": rpms_to_remove}

        for key in self.redis.keys(self.key_pattern):
            assert key != self.key

            task_dict = self.redis.hgetall(key)
            if not task_dict:
                # TOCTOU: The key might no longer exist in DB, even though
                # _our process_ holds the lock!  See the
                # 'self.redis.delete(self.key)' above in check_processed(); the
                # _original process_ removes key without locking, if _third_
                # process commit()ted the status=succeeded.
                # Prior fixing #3770/#3777 we tried to process these deleted
                # tasks and ended up with KeyError on task_dict["task"] below.
                self.log.info("Key %s already processed (by other process), "
                              "removed by the original process (without lock) "
                              "since we listed it.", key)
                continue

            if task_dict.get("status") is not None:
                # skip processed tasks
                self.log.info("Key %s already processed, skip", key)
                continue

            task_opts = json.loads(task_dict["task"])

            # we can process this task!
            self.notify_keys.append(key)

            # append "add" tasks, if that makes sense
            rpms_to_add.update(task_opts["add_content_units"])
            rpms_to_remove.update(task_opts["remove_content_units"])

        # Make sure we are not adding and removing the same packages
        common = rpms_to_add & rpms_to_remove
        rpms_to_add -= common
        rpms_to_remove -= common

        return {"add_content_units": list(rpms_to_add), "remove_content_units": list(rpms_to_remove)}

    def commit(self):
        """
        Report that we processed other createrepo requests.  We don't report
        about failures, we rather kindly let the responsible processes to re-try
        the createrepo tasks.  Requires lock!
        """
        if self.noop:
            return

        for key in self.notify_keys:
            self.log.info("Notifying %s that we succeeded", key)
            self.redis.hset(key, "status", "success")


class PulpClient:
    """
    A client for interacting with Pulp API.

    API documentation:
    - https://docs.pulpproject.org/pulp_rpm/restapi.html
    - https://docs.pulpproject.org/pulpcore/restapi.html

    A note regarding PUT vs PATCH:
    - PUT changes all data and therefore all required fields needs to be sent
    - PATCH changes only the data that we are sending

    A lot of the methods require repository, distribution, publication, etc,
    to be the full API endpoint (called "pulp_href"), not simply their name.
    If method argument doesn't have "name" in its name, assume it expects
    pulp_href. It looks like this:
    /pulp/api/v3/publications/rpm/rpm/5e6827db-260f-4a0f-8e22-7f17d6a2b5cc/
    """

    @classmethod
    def create_from_config_file(cls, path=None, log=None, opts=None):
        """
        Create a Pulp client from a standard configuration file that is
        used by the `pulp` CLI tool
        """
        path = os.path.expanduser(path or "~/.config/pulp/cli.toml")
        with open(path, "rb") as fp:
            config = tomllib.load(fp)
        return cls(config["cli"], log, opts)

    def __init__(self, config, log=None, opts=None):
        self.config = config
        self.timeout = 60
        self.log = log or logging.getLogger(__name__)
        self.opts = opts

    @property
    def auth(self):
        """
        https://requests.readthedocs.io/en/latest/user/authentication/
        """
        return (self.config["username"], self.config["password"])

    @property
    def cert(self):
        """
        See Client Side Certificates
        https://docs.python-requests.org/en/latest/user/advanced/
        """
        return (self.config["cert"], self.config["key"])

    def url(self, endpoint):
        """
        A fully qualified URL for a given API endpoint
        """
        domain = self.config["domain"]
        if domain == "default":
            domain = ""

        relative = os.path.normpath("/".join([
            self.config["api_root"],
            domain,
            endpoint,
        ]))

        # Normpath removes the trailing slash. If it was there, put it back
        if endpoint[-1] == "/":
            relative += "/"
        return self.config["base_url"] + relative

    def send(self, method, url, data=None, files=None, headers=None, timeout=None):
        """
        Performs a "safe request", meaning that if a request fails, we wait
        and try again, and again, until it succeeds.
        """
        # pylint: disable=too-many-positional-arguments
        request = SafeRequest(
            log=self.log,
            try_indefinitely=True,
            timeout=timeout or self.timeout,
            user_agent="crc-pulp-client",
        )
        if all(self.cert):
            request.cert = self.cert
        else:
            request.auth = self.auth

        # If only URI was passed, make it a fully qualified URL
        if not url.startswith(("http://", "https://")):
            url = self.url(url)

        response = request.send(
            url=url,
            method=method,
            data=data,
            files=files,
            headers=headers,
        )
        return response

    def create_repository(self, name):
        """
        Create an RPM repository
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_create
        """
        uri = "/api/v3/repositories/rpm/rpm/"
        data = {"name": name, "retain_repo_versions": 1, "retain_package_versions": 5}
        self.log.info("Pulp: create_repository: %s %s", uri, name)
        return self.send("POST", uri, data)

    def get_repository(self, name):
        """
        Get a single RPM repository
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_list
        """
        # There is no endpoint for querying a single repository by its name,
        # even Pulp CLI does this workaround
        uri = "/api/v3/repositories/rpm/rpm/?"
        uri += urlencode({"name": name, "offset": 0, "limit": 1})
        self.log.info("Pulp: get_repository: %s", uri)
        return self.send("GET", uri)

    def get_distribution(self, name):
        """
        Get a single RPM distribution
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Distributions:-Rpm/operation/distributions_rpm_rpm_list
        """
        # There is no endpoint for querying a single repository by its name,
        # even Pulp CLI does this workaround
        uri = "/api/v3/distributions/rpm/rpm/?"
        uri += urlencode({"name": name, "offset": 0, "limit": 1})
        self.log.info("Pulp: get_distribution: %s", uri)
        return self.send("GET", uri)

    def get_task(self, task):
        """
        Get a detailed information about a task
        """
        return self.get_by_href(task)

    def get_by_href(self, href):
        """
        Get a detailed information about an object
        """
        url = self.config["base_url"] + href
        self.log.info("Pulp: get_by_href: %s", href)
        return self.send("GET", url)

    def create_distribution(self, name, repository, basepath=None):
        """
        Create an RPM distribution
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Distributions:-Rpm/operation/distributions_rpm_rpm_create
        """
        uri = "/api/v3/distributions/rpm/rpm/"
        data = {
            "name": name,
            "repository": repository,
            "base_path": basepath or name,
        }
        self.log.info("Pulp: create_distribution: %s %s", uri, data)
        return self.send("POST", uri, data)

    def update_distribution(self, distribution, publication=None, repository=None):
        """
        Update an RPM distribution
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Distributions:-Rpm/operation/distributions_rpm_rpm_update

        This allows us to point a distribution to either a publication or
        a repository. Not both, that doesn't make sense and Pulp would raise
        "Only one of the attributes 'repository' and 'publication' may be used simultaneously."
        """
        if publication and repository:
            raise RuntimeError("Specify either publication or repository")

        url = self.config["base_url"] + distribution
        data = {
            "publication": publication,
            "repository": repository,
        }
        self.log.info("Pulp: updating distribution %s", distribution)
        return self.send("PATCH", url, data)

    def create_publication(self, repository):
        """
        Create an RPM publication
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Publications:-Rpm/operation/publications_rpm_rpm_create
        """
        uri = "/api/v3/publications/rpm/rpm/"
        data = {"repository": repository}
        self.log.info("Pulp: publishing %s %s", uri, repository)
        return self.send("POST", uri, data)

    def get_publication(self, repository):
        """
        Get a single RPM publication
        https://pulpproject.org/pulp_rpm/restapi/#tag/Publications:-Rpm/operation/publications_rpm_rpm_list
        """
        # There is no endpoint for querying a single publication
        uri = "/api/v3/publications/rpm/rpm/?"
        uri += urlencode({"repository": repository, "offset": 0, "limit": 1})
        self.log.info("Pulp: get_publicatoin: %s", uri)
        return self.send("GET", uri)

    def create_content(self, path, labels, timeout=3600):
        """
        Create content for a given artifact
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Content:-Packages/operation/content_rpm_packages_create
        """
        uri = "/api/v3/content/rpm/packages/upload/"
        with open(path, "rb") as fp:
            data = {"pulp_labels": json.dumps(labels)}
            files = {"file": fp}
            self.log.info("Pulp: create_content: %s %s", uri, path)
            package = self.send("POST", uri, data=data, files=files, timeout=timeout)
        return package

    def create_content_chunked(self, path, labels):
        """
        Split a file into chunks, upload them and reassemble them on the Pulp
        side.
        https://pulpproject.org/pulpcore/docs/user/guides/upload-publish/?h=chunked#chunked-uploads
        https://docs.pulpproject.org/pulp_rpm/restapi.html#tag/Content:-Packages/operation/content_rpm_packages_create
        """
        # Send an initial request, saying that we want to chunk-upload a file
        # of a given size
        uri = "/api/v3/uploads/"
        data = {
            "size": os.path.getsize(path),
        }
        response = self.send("POST", uri, data)
        upload_href = response.json()["pulp_href"]
        upload_url = self.config["base_url"] + upload_href

        with open(path, "rb") as fp:
            sha256 = hashlib.file_digest(fp, "sha256")

        # Upload the chunks
        size = 1024 * 1024 * 100
        for i, chunk in enumerate(self._read_chunks(path, size=size)):
            # Figure out the start and end byte of the chunk within the whole
            # file
            start = i * size
            end = start + len(chunk) - 1
            headers = {
                "Content-Range": "bytes {0}-{1}/*".format(start, end),
            }
            files = {
                "file": (
                    # Filename
                    "chunk{0}".format(i),

                    # Content from bytes that we already have in the memory
                    io.BytesIO(chunk),

                    # A generic MIME type for binary files
                    "application/octet-stream",
                ),
            }
            response = self.send(
                "PUT",
                upload_url,
                files=files,
                headers=headers,
            )
            if not response.ok:
                raise RuntimeError(
                    "Failed to upload chunk {0} of {1}: {2}"
                    .format(i, path, response.reason)
                )

        # Send the file request that we want to reassemble the file
        commit_url = self.config["base_url"] + upload_href + "commit/"
        data = {
            "sha256": sha256.hexdigest(),
        }
        response = self.send("POST", commit_url, data=data)

        # We need to wait until the task finishes. This is the disadvantage
        # compared to standard uploads
        task = response.json()["task"]
        response = self.wait_for_finished_task(task)
        data = response.json()
        if not response.ok or data["state"] != "completed":
            self.log.error("Failed to upload chunked %s", path)
            self.log.error("Response %s: %s", response, response.json())
            raise RuntimeError("Failed to upload chunked {0}".format(path))
        self.log.info("Successfully uploaded chunked %s", path)

        resources = data["created_resources"]
        if len(resources) != 1:
            raise RuntimeError("Unexpected number of created files: {0}".format(resources))

        uri = "/api/v3/content/rpm/packages/upload/"
        data = {
            "artifact": resources[0],
            "pulp_labels": json.dumps(labels),
        }
        self.log.info("Pulp: create_content: %s for artifact %s", uri, resources[0])
        response = self.send("POST", uri, data=data)
        self.log.info("Successfully created a content for chunked uploaded %s", path)
        return response

    def _read_chunks(self, path, size=10000):
        """
        Generate chunks of a given file
        """
        with open(path, "rb") as fp:
            while content := fp.read(size):
                yield content

    def modify_repository_content_locked(self, repository, batch):
        """
        Creates a new repository version and a new publication. Executed under lock.
        """
        if batch.check_processed():
            self.log.info("Task processed by other process")
            return True

        # Merge others' tasks with ours (if any).
        data = batch.options()
        if not data["add_content_units"] and not data["remove_content_units"]:
            # No RPMs to add or remove
            return True
        # Make the request to create a new repository version
        response = self.modify_repository_content(repository, **data)
        if not response.ok:
            self.log.error("Failed to create a new repository version for: %s, %s",
                           repository, response.text)
            return False
        task = response.json()["task"]
        response = self.wait_for_finished_task(task)
        data = response.json()
        if response.ok and data["state"] == "completed":
            self.log.info("Successfully modified Pulp repository content %s", repository)
        else:
            self.log.error("Failed to modify Pulp repository content %s", repository)
            self.log.error("Response %s: %s", response, response.json())
            return False

        # Make a request to create a new publication
        response = self.create_publication(repository)
        task = response.json()["task"]
        response = self.wait_for_finished_task(task)
        data = response.json()
        if response.ok and data["state"] == "completed":
            self.log.info("Successfully created Pulp publication of repository %s", repository)
            return True
        self.log.info("Failed to create Pulp publication of repository %s", repository)
        return False

    def try_lock(self, repository, batch):
        """
        Periodically try to acquire the lock, and execute the modify_repository_content_locked() method.
        """

        while True:

            # We don't have fair locking (locks-first => processes-first).  So to
            # avoid potential indefinite waiting (see issue #1423) we check if the
            # task isn't already processed _without_ having the lock.

            if batch.check_processed(delete_if_not=False):
                self.log.info("Task processed by other process (no-lock)")
                return

            try:
                lockdir = os.environ.get(
                    "COPR_TESTSUITE_LOCKPATH", "/var/lock/copr-backend")
                with lock(repository, lockdir=lockdir, timeout=5, log=self.log):
                    if self.modify_repository_content_locked(repository, batch):
                        # While we still hold the lock, notify others we processed their
                        # task.  Note that we do not commit in case of random exceptions
                        # above.
                        batch.commit()
                        self.log.debug("Repository version and Publication created by this process")
                        break

            except LockTimeout:
                continue  # Try again...

            # we never loop, only upon timeout
            assert False

    def modify_repository_content(self, repository, add_content_units, remove_content_units):
        """
        Add and/or remove a list of artifacts to/from a repository
        https://pulpproject.org/pulp_rpm/restapi/#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_modify
        """
        path = os.path.join(repository, "modify/")
        url = self.config["base_url"] + path
        data = {"add_content_units": add_content_units or [], "remove_content_units": remove_content_units or []}
        return self.send("POST", url, data)

    def add_content(self, repository, artifacts):
        """
        Add a list of artifacts to a repository
        https://pulpproject.org/pulp_rpm/restapi/#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_modify
        """
        batch = BatchedAddRemoveContent(repository, artifacts, None, self.opts, self.log)
        # Put our task to Redis DB and allow _others_ to process our
        # own task.  This needs to be run _before_ the lock() call.
        batch.make_request()
        self.try_lock(repository, batch)
        self.log.info("Pulp: add_content: %s (%s)", repository, artifacts)
        return True

    def delete_content(self, repository, artifacts):
        """
        Delete a list of artifacts from a repository
        https://pulpproject.org/pulp_rpm/restapi/#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_modify
        """
        batch = BatchedAddRemoveContent(repository, None, artifacts, self.opts, self.log)
        # Put our task to Redis DB and allow _others_ to process our
        # own task.  This needs to be run _before_ the lock() call.
        batch.make_request()
        self.try_lock(repository, batch)
        self.log.info("Pulp: delete_content: %s (%s)", repository, artifacts)
        return True

    def get_content(self, build_ids, chroot=None, fields=None):
        """
        Get a list of PRNs for RPMs with provided build ids
        https://pulpproject.org/pulp_rpm/restapi/#tag/Content:-Packages/operation/content_rpm_packages_list
        """
        query = "("
        for i, build_id in enumerate(build_ids):
            if i:
                query += " OR "
            query += f"pulp_label_select=\"build_id={build_id}\""
        query += ")"

        if chroot:
            query += f" AND pulp_label_select=\"chroot={chroot}\""

        all_results = []
        offset = 0
        limit = 1000

        while True:
            params = {"q": query, "offset": offset, "limit": limit}
            if fields:
                params["fields"] = ",".join(fields)

            uri = "api/v3/content/rpm/packages/?" + urlencode(params)
            self.log.debug("Pulp: get_content: fetching page (offset=%d)", offset)

            response = self.send("GET", uri)
            response.raise_for_status()
            data = response.json()
            results = data.get("results", [])

            self.log.debug("Pulp: get_content: fetched %d items", len(results))

            all_results.extend(results)

            next_url = data.get("next")
            if not next_url or len(results) == 0:
                break

            offset += limit

        self.log.info("Pulp: get_content: fetched %d items", len(all_results))
        return PaginatedResponse(all_results, response)

    def delete_repository(self, repository):
        """
        Delete an RPM repository
        https://pulpproject.org/pulp_rpm/restapi/#tag/Repositories:-Rpm/operation/repositories_rpm_rpm_delete
        """
        url = self.config["base_url"] + repository
        self.log.info("Pulp: delete_repository: %s", repository)
        return self.send("DELETE", url)

    def delete_distribution(self, distribution):
        """
        Delete an RPM distribution
        https://pulpproject.org/pulp_rpm/restapi/#tag/Distributions:-Rpm/operation/distributions_rpm_rpm_delete
        """
        url = self.config["base_url"] + distribution
        self.log.info("Pulp: delete_distribution: %s", distribution)
        return self.send("DELETE", url)

    def wait_for_finished_task(self, task, timeout=86400):
        """
        Pulp task (e.g. creating a publication) can be running for an
        unpredictably long time. We need to wait until it is finished to know
        what it actually did.
        """
        start = time.time()
        while True:
            self.log.info("Pulp: polling task status: %s", task)
            response = self.get_task(task)
            if not response.ok:
                break
            if response.json()["state"] not in ["waiting", "running"]:
                break
            if time.time() > start + timeout:
                break
            time.sleep(5)
        return response

    def list_distributions(self, prefix):
        """
        Get a list of distributions whose names match a given prefix
        https://pulpproject.org/pulp_rpm/restapi/#tag/Distributions:-Rpm/operation/distributions_rpm_rpm_list
        """
        uri = "api/v3/distributions/rpm/rpm/?"
        uri += urlencode({"name__startswith": prefix})
        self.log.info("Pulp: list_distributions: %s", uri)
        return self.send("GET", uri)

    def set_label(self, href, name, value):
        """
        Set a label on a given object
        https://pulpproject.org/pulp_rpm/restapi/#tag/Content:-Packages/operation/content_rpm_packages_set_label
        """
        uri = href + "set_label/"
        data = {"key": name, "value": value}
        self.log.info("Pulp: set_label: %s", uri)
        return self.send("POST", uri, data)



" test worker_manager.py "

import os
import time
import logging
from unittest.mock import MagicMock, patch

import pytest
from munch import Munch
from copr_common.enums import DefaultActionPriorityEnum

from copr_common.redis_helpers import get_redis_connection
from copr_common.worker_manager import (
    JobQueue,
    WorkerManager,
    PredicateWorkerLimit,
)
from copr_backend.actions import ActionWorkerManager, ActionQueueTask, Action
from copr_backend.worker_manager import BackendQueueTask

WORKDIR = os.path.dirname(__file__)

REDIS_OPTS = Munch(
    redis_db=9,
    redis_port=7777,
)

log = logging.getLogger()
log.setLevel(logging.DEBUG)

# pylint: disable=too-many-instance-attributes,protected-access


class ToyWorkerManager(WorkerManager):
    # pylint: disable=abstract-method
    process_counter = 0
    task_sleep = 0
    started_in_cycle = 0
    expected_terminations_in_cycle = None

    def start_task(self, worker_id, task):
        self.process_counter += 1

        cmd = [
            'python3',
            os.path.join(WORKDIR, 'action-processor.py'),
            self.process_counter,
            repr(task),
            worker_id,
            self.task_sleep,
        ]

        environ = {} # os.environ.copy()
        task_env = getattr(self, 'environ', None)
        if task_env:
           environ.update(task_env)

        start = time.time()
        #subprocess.check_call(list(map(str, cmd)), env=environ)
        retyped_cmd = list(map(str, cmd))
        self.start_daemon_on_background(retyped_cmd, env=environ)
        self.log.debug("starting-on-background-took %s (%s)",
                       time.time() - start, retyped_cmd)
        self.started_in_cycle += 1

    def _clean_daemon_processes(self):
        """
        Check that we are not leaving any zombies behind us
        """
        waited = super()._clean_daemon_processes()
        self.log.debug("cleaned up %s, started %s", waited, self.started_in_cycle)
        if waited != self.started_in_cycle:
            if self.expected_terminations_in_cycle is not None:
                assert self.expected_terminations_in_cycle == waited
                return waited
            assert False
        return waited

    def run(self, *args, **kwargs):
        self.started_in_cycle = 0
        return super().run(*args, **kwargs)

    def finish_task(self, _w_id, _tinfo):
        pass


class ToyActionWorkerManager(ToyWorkerManager, ActionWorkerManager):
    pass


class ToyQueueTask(BackendQueueTask):
    def __init__(self, _id):
        self._id = _id

    @property
    def id(self):
        return self._id

    @property
    def is_odd(self):
        """ return True if the self.id is odd number """
        return bool(int(self.id) % 2)


class TestPrioQueue(object):
    def setup_method(self, method):
        raw_actions = [0, 1, 2, 3, 3, 3, 4, 5, 6, 7, 8, 9]
        self.queue = JobQueue()
        for action in raw_actions:
            self.queue.add_task(action, priority=10)

        # one task re-added with priority
        self.queue.add_task(7, priority=5)

    def get_tasks(self):
        tasks = []
        while True:
            try:
                tasks += [self.queue.pop_task()]
            except:
                break
        return tasks

    def test_queue_order(self):
        assert self.get_tasks() == [7, 0, 1, 2, 3, 4, 5, 6, 8, 9]

    def test_pop_push(self):
        self.queue.pop_task()
        self.queue.add_task(11) # prio 0 by default
        assert self.get_tasks() == [11, 0, 1, 2, 3, 4, 5, 6, 8, 9]

    def test_push_back(self):
        self.queue.pop_task()
        self.queue.pop_task()
        self.queue.add_task(10, priority=10) # put back
        assert self.get_tasks() == [1, 2, 3, 4, 5, 6, 8, 9, 10]

    def test_shift_prio(self):
        self.queue.add_task(9, priority=5) # move front, but after 7
        self.queue.add_task(6) # move forward
        assert self.get_tasks() == [6, 7, 9, 0, 1, 2, 3, 4, 5, 8]


class BaseTestWorkerManager:
    redis = None
    worker_manager = None

    def setup_method(self, method):
        log.setLevel(logging.DEBUG)
        self.setup_redis()
        self.setup_worker_manager()
        self.setup_tasks()

    def setup_redis(self):
        self.redis = get_redis_connection(REDIS_OPTS)
        self.redis.flushall()

    def setup_worker_manager(self):
        self.worker_manager = ToyWorkerManager(
            redis_connection=self.redis,
            max_workers=5,
            log=log)

    def setup_tasks(self, exclude=None):
        """ Fill the task list """
        if exclude is None:
            exclude = []
        raw_actions = [0, 1, 2, 3, 3, 3, 4, 5, 6, 7, 8, 9]
        actions = [ToyQueueTask(action) for action in raw_actions
                   if not action in exclude]
        self.worker_manager.clean_tasks()
        for action in actions:
            self.worker_manager.add_task(action)

    def workers(self):
        return self.worker_manager.worker_ids()

    def remaining_tasks(self):
        count = 0
        while True:
            try:
                self.worker_manager.tasks.pop_task()
                count += 1
            except:
                break
        return count


class TestLimitedWorkerManager(BaseTestWorkerManager):
    limits = []
    def setup_worker_manager(self):
        self.limits = [
            PredicateWorkerLimit(lambda x: x.is_odd, 3, name="odd"),
            PredicateWorkerLimit(lambda x: not x.is_odd, 2, name="even"),
        ]
        self.worker_manager = ToyWorkerManager(
            redis_connection=self.redis,
            max_workers=50,
            log=log,
            limits=self.limits)

    @patch('copr_common.worker_manager.time.sleep')
    @patch('copr_common.worker_manager.time.time')
    def test_that_limits_are_respected(self, mc_time, _mc_sleep, caplog):
        # each time.time() call incremented by 1
        self.worker_manager.task_sleep = 5
        self.worker_manager.worker_timeout_start = 1000
        mc_time.side_effect = range(1000)
        self.worker_manager.run(timeout=150)
        messages = [
            "Task '4' skipped, limit info: 'even', "
            "matching: worker:0, worker:2",
            "Task '6' skipped, limit info: 'even', "
            "matching: worker:0, worker:2",
            "Task '7' skipped, limit info: 'odd', "
            "matching: worker:1, worker:3, worker:5",
            "Task '8' skipped, limit info: 'even', "
            "matching: worker:0, worker:2",
            "Task '9' skipped, limit info: 'odd', "
            "matching: worker:1, worker:3, worker:5",
        ]
        for msg in messages:
            assert ('root', logging.DEBUG, msg) in caplog.record_tuples

        # Even though the "even" limit kicked-out task 4, the task 5 is still
        # successfully started because that's the third "odd" task.  The rest of
        # tasks is just skipped.
        assert ('root', logging.INFO,
                "Starting worker worker:5, task.priority=0") in \
            caplog.record_tuples

        # finish the task now
        self.redis.hset("worker:5", "status", "0")

        self.setup_tasks()  # re-calculate limits
        self.worker_manager.run(timeout=150)

        # check worker manager recognized the finished task
        assert ('root', logging.INFO, "Finished worker worker:5") \
                in caplog.record_tuples

        # worker 7 is not yet started, it waits for next run() because run()
        # doesn't free the limit quota when removing worker from Redis (see
        # issue #1415)
        worker_7_started = "Starting worker worker:7, task.priority=0"
        assert ('root', logging.INFO, worker_7_started) not in \
            caplog.record_tuples

        # another run finally takes the task 7, because 5 is done
        self.setup_tasks(exclude=[5])
        self.worker_manager.run(timeout=150)
        assert ('root', logging.INFO, worker_7_started) in \
            caplog.record_tuples


class TestWorkerManager(BaseTestWorkerManager):
    def test_worker_starts(self):
        task = self.worker_manager.tasks.pop_task()
        assert task.id == 0
        self.worker_manager._start_worker(task, time.time())
        worker_id = self.worker_manager.get_worker_id(repr(task))
        assert len(self.redis.keys(worker_id)) == 1
        self.worker_manager._clean_daemon_processes()

    def test_number_of_tasks(self):
        assert self.remaining_tasks() == 10

    def test_task_to_worker_id(self):
        wid = "{}:123".format(self.worker_manager.worker_prefix)
        assert self.worker_manager.get_task_id_from_worker_id(wid) == "123"

    @patch('copr_common.worker_manager.time.sleep')
    def test_preexisting_broken_worker(self, _mc_sleep, caplog):
        """ from previous systemctl restart """
        fake_worker_name = self.worker_manager.worker_prefix + ":fake"
        self.redis.hset(fake_worker_name, "foo", "bar")
        self.worker_manager.run(timeout=0.0001)
        msg = "Missing 'allocated' flag for worker " + fake_worker_name
        assert ('root', logging.INFO, msg) in caplog.record_tuples

    def test_cancel_task(self):
        self.redis.hset('worker:4', 'allocated', 1)
        self.worker_manager.cancel_task_id(3)
        self.worker_manager.cancel_task_id(4)
        self.worker_manager.cancel_task_id(666)
        assert self.redis.hgetall('worker:3') == {}
        assert "cancel_request" in self.redis.hgetall('worker:4')

    def test_slow_priority_queue_filling(self):
        """
        We discovered that adding tasks to a priority queue was a bottleneck
        when having a large (70k+ builds) queue, see #2095. Make sure this never
        happen again.
        """
        tasks = [ToyQueueTask(i) for i in range(100000)]

        # We need to run this test with logging only INFO, otherwise we waste
        # around 5 seconds just on running self.log.debug because we need to
        # connect to redis for each call
        # The point of this test is to make sure that adding tasks to priority
        # queue is not a bottleneck on production, and we don't use DEBUG there
        # anyway.
        log.setLevel(logging.INFO)

        t1 = time.time()
        for task in tasks:
            self.worker_manager.add_task(task)
        t2 = time.time()

        # It should actually be faster than 1 second but I am adding one to
        # prevent false alarms in case somebody has a slow machine
        assert t2 - t1 < 2


def wait_pid_exit(pid):
    """ wait till pid stops responding to no-op kill 0 """
    while True:
        try:
            os.kill(int(pid), 0)
        except OSError:
            return True
        time.sleep(0.1)


class TestActionWorkerManager(BaseTestWorkerManager):
    # pylint: disable=attribute-defined-outside-init
    def setup_worker_manager(self):
        self.worker_manager = ToyActionWorkerManager(
            redis_connection=self.redis,
            max_workers=5,
            log=log)

        prefix = 'toy:' + str(time.time())
        self.worker_manager.worker_prefix = prefix
        prefix += ':'
        self.wprefix = prefix
        self.w0 = prefix + '0'
        self.w1 = prefix + '1'
        self.worker_manager.frontend_client = MagicMock()

    def setup_tasks(self, exclude=None):
        if exclude is None:
            exclude = []
        raw_actions = [x for x in [0, 1, 2, 3, 3, 3, 4, 5, 6, 7, 8, 9]
                       if x not in exclude]
        self.worker_manager.clean_tasks()
        for action in raw_actions:
            action = ToyQueueTask(action)
            self.worker_manager.add_task(action)

    def test_run_starts_the_workers(self):
        self.worker_manager.run(timeout=0.0001)
        workers = self.workers()
        assert len(workers) == 1
        assert workers[0] == self.w0

        args = self.wait_field(self.w0, 'status')
        assert 'status' in args
        assert 'PID' in args
        assert 'started' in args

        self.worker_manager.run(timeout=0.0001)

        keys = self.workers()
        assert self.w0 not in keys
        # we are not sure 'toy:1' had a chance to start
        assert len(keys) <= 1

    def test_delete_not_allocated_workers(self):
        self.worker_manager.run(timeout=0.0001)
        assert self.w0 in self.workers()
        # "allocated" can be the only one hash attribute, so add another one to
        # not loose the key from redis DB
        self.redis.hset(self.w0, "keep_me_around", "yes?")
        self.redis.hdel(self.w0, "allocated")
        self.worker_manager.run(timeout=0.0001)
        params = self.redis.hgetall(self.w0)
        if params:
            # The action-processor.py script re-added itself to the database,
            # but still the 'keep_me_around' flag disappeared after worker
            # manager removed it.
            assert "keep_me_around" not in params

    def test_delete_not_started_workers(self):
        self.worker_manager.environ = {'FAIL_EARLY': '1'}
        self.worker_manager.worker_timeout_start = 0
        self.worker_manager.run(timeout=0.0001)
        assert self.workers() == [self.w0]
        self.worker_manager.run(timeout=0.0001)
        # toy 0 is deleted now
        assert self.workers() == [self.w1]

    def wait_field(self, worker, field):
        for _ in range(0, 100):
            time.sleep(0.25)
            params = self.redis.hgetall(self.w0)
            if field in params:
                return params
        raise Exception("Unsuccessful wait for {} in {}".format(worker, field))

    @pytest.mark.parametrize('fail', ['FAIL_STARTED_PID', 'FAIL_STARTED'])
    @patch('copr_common.worker_manager.time.time')
    def test_delete_not_finished_workers(self, mc_time, fail):
        self.worker_manager.environ = {fail: '1'}
        self.worker_manager.worker_timeout_deadcheck = 0.4

        # each time.time() call incremented by 1
        mc_time.side_effect = range(1000)

        # first loop just starts the toy:0 worker
        with patch('copr_common.worker_manager.time.sleep'):
            self.worker_manager.run(timeout=1)

        params = self.wait_field(self.w0, 'started')
        assert self.w0 in self.workers()
        assert 'started' in params

        if fail == 'FAIL_STARTED':
            # make sure kernel cleans up the process, so the next wm.run()
            # certainly sets the 'delete' flag
            wait_pid_exit(params['PID'])

        # toy 0 is marked for deleting
        with patch('copr_common.worker_manager.time.sleep'):
            self.worker_manager.run(timeout=1)
        assert 'delete' in self.redis.hgetall(self.w0)

        # toy 0 should be deleted
        with patch('copr_common.worker_manager.time.sleep'):
            self.worker_manager.run(timeout=1)
        keys = self.workers()
        assert self.w1 in keys
        assert self.w0 not in keys

    def test_all_passed(self, caplog):
        # It is a lot of fun with Popen().  It seems it has some zombie reaping
        # mechanism.  If the calling function objects are destroyed (including
        # the Popen() return value reference), the future call to Popen() seems
        # to just reap the old Popen() processes.
        self.worker_manager.expected_terminations_in_cycle = 5

        self.worker_manager.run(timeout=100)
        for i in range(0, 10):
            smsg = "Starting worker {}{}, task.priority=0"
            assert ('root', 20, smsg.format(self.wprefix, i)) in \
                caplog.record_tuples
            assert ('root', 20, 'Finished worker {}{}'.format(self.wprefix, i)) in caplog.record_tuples

    def test_add_task_for_running_worker(self, caplog):
        # at least 'toy:0' should be reeady
        self.worker_manager.run(timeout=0.0001)

        queue_len = len(self.worker_manager.tasks.prio_queue)
        self.worker_manager.add_task(ToyQueueTask(0))
        assert len(self.worker_manager.tasks.prio_queue) == queue_len
        assert ('root', logging.DEBUG,
                "Task 0 already has a worker process") in caplog.record_tuples

    def test_empty_queue_but_workers_running(self):
        'check that sleep(1) is done if queue is empty, but some workers exist'

        self.worker_manager.clean_tasks()

        # only one task, but it will take some time.
        self.worker_manager.task_sleep = 0.5
        self.worker_manager.add_task(ToyQueueTask(0))

        # start the worker
        self.worker_manager.run(timeout=0.0001) # start them task

        with patch('copr_common.worker_manager.time.sleep') as sleep:
            # we can spawn more workers, but queue is empty
            self.worker_manager.run(timeout=0.0001)
            assert sleep.called
        assert len(self.worker_manager.worker_ids()) == 1

        # let the task finish
        self.wait_field(self.w0, 'status')

        # check that we don't sleep here (no worker, no task)
        with patch('copr_common.worker_manager.time.sleep') as sleep:
            self.worker_manager.run(timeout=0.0001)
            assert not sleep.called

        assert len(self.worker_manager.worker_ids()) == 0

    def test_that_we_check_aliveness(self):
        """
        Worker Manager checks whether worker is running each 'worker_timeout_deadcheck'
        period, check whether it works
        """
        self.worker_manager.task_sleep = 3 # assure task takes some time
        self.worker_manager.clean_tasks()
        self.worker_manager.add_task(ToyQueueTask(0))
        self.worker_manager.worker_timeout_start = 1
        self.worker_manager.worker_timeout_deadcheck = 1.5

        # start the worker
        self.worker_manager.run(timeout=0.0001)

        # let the task start
        self.wait_field(self.w0, 'PID')

        # timeout for liveness check not yet left
        self.worker_manager.run(timeout=0.0001)
        params = self.redis.hgetall(self.w0)
        if 'checked' in params:
            # slow builder, the delay between previous two run() calls were so
            # long so the second one managed to even check whether the worker is
            # alive.  So if that happened, the delay needs to be larger than
            # deadcheck at least.
            checked = float(params['checked'])
            started = float(params['allocated'])
            assert started + self.worker_manager.worker_timeout_deadcheck <= checked

        # time for check..
        time.sleep(1.5)
        self.worker_manager.run(timeout=0.0001)
        params = self.redis.hgetall(self.w0)
        assert 'checked' in params

        # let the task finish
        self.wait_field(self.w0, 'status')
        self.worker_manager.run(timeout=0.0001)

        assert len(self.worker_manager.worker_ids()) == 0

    def test_max_workers_has_effect(self):
        self.worker_manager.max_workers = 1
        self.worker_manager.run(timeout=1)
        assert self.w0 in self.workers()
        assert self.w1 not in self.workers()


class TestActionWorkerManagerPriorities(BaseTestWorkerManager):
    def setup_worker_manager(self):
        self.worker_manager = ToyActionWorkerManager(
            redis_connection=self.redis,
            max_workers=5,
            log=log)

    def setup_tasks(self, exclude=None):
        _unused = (self, exclude)

    def pop(self):
        return self.worker_manager.tasks.pop_task()

    def test_actions_priorities(self):
        frontend_data = [
            {"id": 10, "priority": DefaultActionPriorityEnum("delete")},
            {"id": 11, "priority": DefaultActionPriorityEnum("delete")},
            {"id": 12, "priority": DefaultActionPriorityEnum("createrepo")},
            {"id": 13, "priority": DefaultActionPriorityEnum("update_comps")},
            {"id": 14, "priority": DefaultActionPriorityEnum("rawhide_to_release")},
            {"id": 15, "priority": DefaultActionPriorityEnum("rawhide_to_release")},
            {"id": 16, "priority": DefaultActionPriorityEnum("build_module")},
            {"id": 17, "priority": DefaultActionPriorityEnum("cancel_build")},
            {"id": 18, "priority": DefaultActionPriorityEnum("fork")},
            {"id": 19, "priority": DefaultActionPriorityEnum("gen_gpg_key")},
            {"id": 20, "priority": 0},
        ]
        for action in frontend_data:
            queue_task = ActionQueueTask(Action(MagicMock(), action, log=log))
            self.worker_manager.add_task(queue_task)

        assert self.pop().id == 19
        assert self.pop().id == 17

        # These have the same priority and the queue is FIFO
        assert self.pop().id == 12
        assert self.pop().id == 13
        assert self.pop().id == 16
        assert self.pop().id == 18
        assert self.pop().id == 20

        assert self.pop().id == 10
        assert self.pop().id == 11

        assert self.pop().id == 14
        assert self.pop().id == 15

        # Tasks queue is empty now
        with pytest.raises(KeyError) as ex:
            self.pop()
        assert "empty" in str(ex)

    def test_backend_priority_adjustments(self):
        """
        Test that backend still can adjust or ultimately override priorities
        """
        frontend_data = [
            {"id": 10, "priority": DefaultActionPriorityEnum("delete")},
            {"id": 11, "priority": DefaultActionPriorityEnum("delete")},
            {"id": 12, "priority": DefaultActionPriorityEnum("createrepo")},
            {"id": 13, "priority": DefaultActionPriorityEnum("gen_gpg_key")},
            {"id": 14, "priority": DefaultActionPriorityEnum("fork")},
        ]
        actions = [ActionQueueTask(Action(MagicMock(), action, log)) for action in frontend_data]

        # QueueTask.backend_priority is a property which should be
        # overriden when in the class descendants.
        delattr(BackendQueueTask, "backend_priority")
        actions[0].backend_priority = 0
        actions[1].backend_priority = -999
        actions[2].backend_priority = 999
        actions[3].backend_priority = 0
        actions[4].backend_priority = -900

        for action in actions:
            self.worker_manager.add_task(action)

        assert self.pop().id == 11
        assert self.pop().id == 14
        assert self.pop().id == 13
        assert self.pop().id == 10
        assert self.pop().id == 12



#!/usr/bin/python

from setuptools import setup, find_packages

long_description = """\
COPR is lightweight build system. It allows you to create new project in WebUI,
and submit new builds and COPR will create yum repository from latest builds.

This package contains backend."""

__description__ = "Python client for copr service."
__author__ = "Copr Team"
__author_email__ = "copr-team@redhat.com"
__url__ = "https://github.com/fedora-copr/copr"


setup(
    name='copr-backend',
    version="2.11",
    description=__description__,
    long_description=long_description,
    author=__author__,
    author_email=__author_email__,
    url=__url__,
    license='GPLv2+',
    packages=find_packages(exclude=('tests*',)),
    package_data={'': ['*.j2']},
    include_package_data=True,
    zip_safe=False,
)



"""
Allocate VMs
"""

import time
import yaml
from resalloc.client import Connection as ResallocConnection


class RemoteHostAllocationTerminated(Exception):
    """ raised when something happened during VM allocation """


class RemoteHost:
    """
    Remote host allowing us to ssh.
    """
    hostname = None
    ssh_port = None

    _is_ready = False

    def check_ready(self):
        """
        Check that the host is ready (usually it means that it is ready to talk
        over ssh layer).
        :raises RemoteHostAllocationTerminated exception.
        """
        raise NotImplementedError

    def release(self):
        """
        Declare that we don't need this host anymore.  It is up to the VM
        provider what it will do with this host.
        """
        raise NotImplementedError

    def _check_ready(self):
        if self._is_ready:
            return True
        self._is_ready = self.check_ready()
        return self._is_ready

    @property
    def info(self):
        """ user-readable info about the host """
        return "no info"

    @staticmethod
    def _incremental_sleep(sleeptimes=None):
        """
        Wait cca two minutes with faster cadence, and then fallback to one
        minute sleep.
        """
        counter = 0
        if sleeptimes is None:
            # wait cca two minutes, then fallback to 60s sleep(s)
            sleeptimes = [3, 3, 6, 6, 10, 10, 10, 20, 20, 20, 30]

        while True:
            try:
                sleeptime = sleeptimes[counter]
            except IndexError:
                sleeptime = 60
            time.sleep(sleeptime)
            yield
            counter += 1

    def wait_ready(self):
        """
        Passively wait till self.is_ready is true.  Return True if we
        successfully waited for the VM.
        """
        sleep = self._incremental_sleep()
        try:
            while True:
                if self._check_ready():
                    return True
                next(sleep)
        except RemoteHostAllocationTerminated:
            pass
        return False


class ResallocHost(RemoteHost):
    """
    VM representation, and box allocation mechanism using pre-configured
    resalloc server:
    https://github.com/praiskup/resalloc
    """
    ticket = None
    name = None
    requested_tags = None

    def parse_ticket_data(self):
        """
        Historically we expected a single-line input containing hostname/IP.  We
        continue to support this format.  But if the output looks like YAML
        document, we parse the output and detect additional metadata.
        """
        output = str(self.ticket.output)
        lines = output.split("\n")
        if lines[0] == "---":
            try:
                data = yaml.safe_load(output)
                # We expect IP or hostname here
                self.hostname = data["host"]
                # RESALLOC_NAME
                self.name = data["name"]
                if 'ssh_port' in data:
                    self.ssh_port = data["ssh_port"]
            except yaml.YAMLError as exc:
                raise RemoteHostAllocationTerminated(
                    f"Can't parse YAML data from the resalloc ticket:\n{output}") from exc
            except KeyError as exc:
                raise RemoteHostAllocationTerminated(
                    f"Missing YAML fields in the resalloc ticket output\n{output}") from exc
        else:
            # The old format
            self.hostname = lines[0]

    def check_ready(self):
        self.ticket.collect()
        if self.ticket.closed:
            # canceled, or someone else closed the ticket
            raise RemoteHostAllocationTerminated
        if not self.ticket.ready:
            return False
        self.parse_ticket_data()
        return True

    def release(self):
        self.ticket.close()

    @property
    def info(self):
        message = "ResallocHost"
        if self.ticket:
            message += ", ticket_id={}".format(self.ticket.id)
        if self.hostname:
            message += ", hostname={}".format(self.hostname)
        if self.name:
            message += ", name={}".format(self.name)
        if self.requested_tags:
            message += f", requested_tags={self.requested_tags}"
        if self.ssh_port is not None:
            message += f", port={self.ssh_port}"
        return message


class HostFactory:
    """ Abstract host provider """
    def get_host(self, tags=None, sandbox=None):
        """
        Return new box instance which is not yet ready.  It is up to the caller
        to wait.
        """
        raise NotImplementedError


class ResallocHostFactory(HostFactory):
    """
    Provide worker hosts using resalloc server:
    https://github.com/praiskup/resalloc
    """
    def __init__(self, server="http://localhost:49100"):
        self.conn = ResallocConnection(
            server, request_survives_server_restart=True)

    def get_host(self, tags=None, sandbox=None):
        request_tags = ["copr_builder"]
        if tags:
            request_tags.extend(tags)

        host = ResallocHost()
        host.ticket = self.conn.newTicket(request_tags, sandbox)
        host.requested_tags = request_tags
        return host



"""
Test classes which inherit from WorkerLimit
"""

# pylint: disable=protected-access

from copr_common.worker_manager import (
    HashWorkerLimit,
    PredicateWorkerLimit,
    StringCounter,
)
from copr_backend.worker_manager import BackendQueueTask
from copr_backend.rpm_builds import (
    ArchitectureWorkerLimit,
    ArchitectureUserWorkerLimit,
    BuildTagLimit,
    BuildQueueTask,
)

TASKS = [{
    "build_id": 7,
    "task_id": "7",
    "project_owner": "cecil",
}, {
    "build_id": 7,
    "task_id": "7-fedora-rawhide-x86_64",
    "chroot": "fedora-rawhide-x86_64",
    "project_owner": "cecil",
    "sandbox": "sb1",
}, {
    "build_id": 4,
    "task_id": "7-fedora-32-x86_64",
    "chroot": "fedora-32-x86_64",
    "project_owner": "bedrich",
    "sandbox": "sb2",
}, {
    "build_id": 4,
    "task_id": "7-fedora-31-x86_64",
    "chroot": "fedora-31-x86_64",
    "project_owner": "bedrich",
    "sandbox": "sb2",
    "tags": ["special_requirement"],
}, {
    "build_id": 4,
    "task_id": "7-fedora-31-aarch64",
    "chroot": "fedora-31-aarch64",
    "project_owner": "bedrich",
    "sandbox": "sb2",
}]

class _QT(BackendQueueTask):
    def __init__(self, _id):
        self._id = _id

    @property
    def id(self):
        return self._id

    @property
    def always_true(self):
        """ this is for predicate tests """
        return True

    @property
    def sometimes_true(self):
        """ this is for predicate tests """
        return bool(int(self._id) % 2)

    @property
    def group(self):
        """ assign task to group_{0,1,2} per modulo operator """
        return "group_{}".format(int(self.id) % 3)

def test_predicate_worker_limit():
    # pylint: disable=protected-access
    wl = PredicateWorkerLimit(lambda x: x.always_true, 3)
    wl.worker_added("1", _QT(1))
    wl.worker_added("2", _QT(2))
    assert wl.check(_QT(3))
    wl.worker_added("3", _QT(3))
    assert wl.check(_QT(4)) is False
    wl.clear()
    assert wl.check(_QT(4))
    wl.worker_added("4", _QT(4))
    # check memory leaks
    assert wl._refs.keys() == set(["4"])
    wl.worker_added("3", _QT(3))
    assert set(wl._refs.keys()) == set(["3", "4"])
    wl.clear()
    assert set(wl._refs.keys()) == set([])

def test_predicate_worker_limit_sometimes():
    # pylint: disable=protected-access
    wl = PredicateWorkerLimit(lambda x: x.sometimes_true, 2)
    wl.worker_added("0", _QT(0))
    wl.worker_added("1", _QT(1))
    assert wl.check(_QT(3))
    wl.worker_added("3", _QT(3))
    assert wl.check(_QT(5)) is False

def test_group_worker_limit():
    wl = HashWorkerLimit(lambda x: x.group, 2)
    for _ in ["first", "cleared", "cleared"]:
        for task in [0, 1, 2]:
            wl.worker_added(str(task), _QT(str(task)))

        for task in [3, 4, 5]:
            qt = _QT(str(task))
            assert wl.check(qt)
            wl.worker_added(str(task), qt)

        qt = _QT("6")
        assert not wl.check(qt)  # limit raised
        wl.clear()

    # check mem leaks
    assert wl._groups._counter == {}
    assert wl._refs == {}

def test_worker_limit_info():
    limits = [
        PredicateWorkerLimit(lambda _: True, 8),
        PredicateWorkerLimit(lambda _: True, 8, name='allmatch'),
        HashWorkerLimit(lambda x: x.owner, 4),
        HashWorkerLimit(lambda x: x.sandbox, 2, name='sandbox'),
        ArchitectureWorkerLimit("x86_64", 3),
        ArchitectureWorkerLimit("aarch64", 2),
        ArchitectureUserWorkerLimit("aarch64", 2),
        BuildTagLimit("special_requirement", 1),
    ]
    tasks = [BuildQueueTask(t) for t in TASKS]
    for limit in limits:
        for task in tasks:
            limit.worker_added("w:" + str(task.id), task)
    assert ["limit info: " + limit.info() for limit in limits] == [
        "limit info: Unnamed 'PredicateWorkerLimit' limit, matching: w:7, "
        'w:7-fedora-rawhide-x86_64, w:7-fedora-32-x86_64, w:7-fedora-31-x86_64, '
        'w:7-fedora-31-aarch64',
        "limit info: 'allmatch', matching: w:7, w:7-fedora-rawhide-x86_64, "
        'w:7-fedora-32-x86_64, w:7-fedora-31-x86_64, w:7-fedora-31-aarch64',
        "limit info: Unnamed 'HashWorkerLimit' limit, counter: cecil=2, bedrich=3",
        "limit info: 'sandbox', counter: sb1=1, sb2=3",
        "limit info: 'arch_x86_64', matching: w:7-fedora-rawhide-x86_64, w:7-fedora-32-x86_64, w:7-fedora-31-x86_64",
        "limit info: 'arch_aarch64', matching: w:7-fedora-31-aarch64",
        "limit info: 'arch_aarch64_owner', counter: aarch64_bedrich=1",
        "limit info: 'tag_special_requirement', matching: w:7-fedora-31-x86_64",
    ]

def test_string_counter():
    counter = StringCounter()
    counter2 = StringCounter()
    counter.add(None)
    counter2.add(None)
    assert str(counter) == ""
    counter.add("foo")
    counter.add("bar")
    counter.add("foo")
    counter2.add("baz")
    assert str(counter) == "foo=2, bar=1"
    assert str(counter2) == "baz=1"



"""
Singleton-like objects to simplify ubiquitous Copr Backend configuration.
"""

from copr_backend.app import App

# the application
app = App()

# configuration, usually in /etc/copr/copr-be.conf
config = app.opts

# the default logger
log = app.log



"""
Test the ./run/copr-backend-unknown-resalloc-tickets.py script
"""

import contextlib
import importlib
import io
from unittest import mock

from testlib import patch_path_run, patch_path_fake_executables, patch_getpwuid

SCRIPT_NAME = 'copr-backend-unknown-resalloc-tickets.py'
SPEC = importlib.util.spec_from_file_location("spec", "run/" + SCRIPT_NAME)
SCRIPT = importlib.util.module_from_spec(SPEC)
SPEC.loader.exec_module(SCRIPT)

EXPECTED_TICKETS = {5273992, 5273994, 5273997, 5273999, 5274000, 5274329,
                    5274764, 5274768, 5274772, 5274773, 5274774, 5274775}


def main_output(ids):
    with mock.patch.object(SCRIPT, "all_ids", return_value=ids):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            SCRIPT._main()  # pylint: disable=protected-access
            return f.getvalue()


@patch_path_run
@patch_path_fake_executables("unknown_tickets")
@patch_getpwuid("resalloc")
def test_unknown_tickets():
    # not ticket found in DB
    assert main_output({}) == "no tickets taken\n"
    assert SCRIPT.used_ids() == EXPECTED_TICKETS

    # no unused ticket
    assert main_output(EXPECTED_TICKETS) == ""

    assert main_output(EXPECTED_TICKETS | {1}) == """\
These are old tickets, Copr only uses newer tickets, close them:
resalloc ticket-close 1
"""
    assert main_output(EXPECTED_TICKETS | {6000000}) == """\
These tickets are relatively new for closing blindly, double check!
resalloc ticket-check 6000000
"""
    assert main_output(EXPECTED_TICKETS | {1, 6000000}) == """\
resalloc ticket-close 1
resalloc ticket-check 6000000
"""



"""
Abstract class Dispatcher for Build/Action dispatchers.
"""

from copr_common.dispatcher import Dispatcher
from copr_backend.frontend import FrontendClient
from copr_backend.helpers import get_redis_logger


class BackendDispatcher(Dispatcher):
    """
    Base class for backend dispatchers
    """

    # This is still a base class, abstract methods will be overriden
    # in its descendants
    # pylint: disable=abstract-method

    def __init__(self, opts):
        super().__init__(opts)

        logger_name = 'backend.{}_dispatcher'.format(self.task_type)
        logger_redis_who = '{}_dispatcher'.format(self.task_type)
        self.log = get_redis_logger(self.opts, logger_name, logger_redis_who)

        self.frontend_client = FrontendClient(self.opts, self.log)
        self.sleeptime = opts.sleeptime



import copy
import os

from urllib.parse import urljoin
from copr_backend.exceptions import CoprBackendSrpmError
from copr_backend.helpers import build_target_dir


class BuildJob(object):
    def __init__(self, task_data, worker_opts):
        """
            Creates build job object
            :param dict task_dict: dictionary with the following fields
                (based frontend.models.Build)::

                - pkgs: list of space separated urls of packages to build
                - timeout: maximum allowed time of build, build will fail if exceeded # unused
                - project_owner:
                - project_name:
                - project_dirname:
                - submitter:

            :param dict worker_opts: worker options, fields::

                - destdir: worker root directory to store results
                - results_baseurl: root url to stored results
                - timeout: default worker timeout

        """
        # pylint: disable=too-many-statements

        self.timeout = worker_opts.timeout
        self.frontend_base_url = worker_opts.frontend_base_url
        self.memory_reqs = None
        self.enable_net = True

        self.project_owner = None
        self.project_name = None
        self.project_dirname = None
        self.submitter = None

        self.ended_on = None
        self.started_on = None
        self.submitted_on = None

        self.status = None
        self.chroot = None
        self.arch = None # parsed from chroot

        self.buildroot_pkgs = None

        self.task_id = None
        self.build_id = None

        self.package_name = None
        self.package_version = None

        self.git_repo = None
        self.git_hash = None
        self.git_branch = None

        self.source_type = None
        self.source_json = None

        self.pkg_name = None
        self.pkg_main_version = None
        self.pkg_epoch = None
        self.pkg_release = None

        self.srpm_url = None
        self.uses_devel_repo = None
        self.sandbox = None

        self.results = None
        self.appstream = None
        self.allow_user_ssh = None
        self.ssh_public_keys = None
        self.storage = None
        self.repos = []

        # TODO: validate update data
        for key, val in task_data.items():
            key = str(key)
            setattr(self, key, val)

        if self.chroot:
            self.arch = self.chroot.rsplit("-", 2)[2]

        if str(self.task_id) == str(self.build_id):
            self.chroot = 'srpm-builds'

        if task_data["appstream"]:
            self.appstream = task_data["appstream"]

        self.destdir = os.path.normpath(os.path.join(
            worker_opts.destdir,
            task_data["project_owner"],
            task_data["project_dirname"]
        ))

        self.results_repo_url = os.path.join(
            worker_opts.results_baseurl,
            task_data["project_owner"],
            task_data["project_dirname"],
        )

        # TODO: We should rename this attribute.  This one is used by Frontend
        # to store updated "target_dir_name" to BuildChroot database.  But the
        # name is terrible, and clashes with self.results_dir (plural).
        self.result_dir = self.target_dir_name

        self.built_packages = ""

        self.tags = ["arch_{}".format(self.arch if self.arch else "noarch")]
        if "tags" in task_data:
            self.tags.extend(task_data["tags"])

    @property
    def results_dir_url(self):
        return os.path.join(self.results_repo_url, self.chroot, self.target_dir_name)

    @property
    def chroot_dir(self):
        return os.path.normpath("{}/{}".format(self.destdir, self.chroot))

    @property
    def results_dir(self):
        return os.path.join(self.chroot_dir, self.target_dir_name)

    @property
    def target_dir_name(self):
        if self.chroot == "srpm-builds":
            return build_target_dir(self.build_id, None)

        return build_target_dir(self.build_id, self.package_name)

    @property
    def backend_log(self):
        """
        The log file which is "live" appended to build resultdir by copr
        backend background process.
        """
        return os.path.join(self.results_dir, "backend.log")

    @property
    def builder_log(self):
        """
        The live log continuously transferred from builder.
        """
        return os.path.join(self.results_dir, "builder-live.log")

    @property
    def review_log(self):
        """
        The log file from running fedora-review
        """
        return os.path.join(self.results_dir, "fedora-review/fedora-review.log")

    @property
    def rsync_log_name(self):
        return "build-{:08d}.rsync.log".format(self.build_id)

    def update(self, data_dict):
        """

        :param dict data_dict:
        """
        # TODO: validate update data
        self.__dict__.update(data_dict)

    def validate(self):
        """
        Make sure the build results don't contain anything problematic
        """
        if self.pkg_name and len(self.pkg_name) > 100:
            msg = "Too long package name: {0}".format(self.pkg_name)
            # Truncate the package name otherwise frontend won't be able to
            # handle it
            self.pkg_name = None
            raise CoprBackendSrpmError(msg)

    def to_dict(self):
        """

        :return dict: dictified build job
        """
        result = copy.deepcopy(self.__dict__)
        result["id"] = self.build_id
        result["mockchain_macros"] = self.mockchain_macros
        return result

    @property
    def mockchain_macros(self):
        return {
            "copr_username": self.project_owner,
            "copr_projectname": self.project_name,
            "vendor": "Fedora Project COPR ({0}/{1})".format(
                self.project_owner, self.project_name)
        }

    @property
    def pkg_version(self):
        """
        Canonical version presentation release and epoch
        "{epoch}:{version}-{release}"
        """
        if self.pkg_main_version is None:
            return None
        if self.pkg_epoch:
            full_version = "{}:{}".format(self.pkg_epoch, self.pkg_main_version)
        else:
            full_version = "{}".format(self.pkg_main_version)

        if self.pkg_release:
            full_version += "-{}".format(self.pkg_release)
        return full_version

    def __str__(self):
        return str(self.__unicode__())

    def __unicode__(self):
        return u"BuildJob<id: {build_id}, owner: {project_owner}, project: {project_name}, project_dir: {project_dirname}" \
               u"git branch: {git_branch}, git hash: {git_hash}, status: {status} >".format(**self.__dict__)

    @property
    def took_seconds(self):
        """ Number of seconds spent on building this package """
        if self.ended_on is None or self.started_on is None:
            return None
        return self.ended_on - self.started_on

    @property
    def task_url(self):
        """
        URL where a builder can downloading the JSON definition for this task
        """
        base = "/backend/get-build-task/"
        if self.chroot == "srpm-builds":
            base = "/backend/get-srpm-build-task/"
        path = urljoin(base, self.task_id)
        return urljoin(self.frontend_base_url, path)
